###############################################################################
#
# PDBtoXLSX - A tool for converting a WinLogStats PDB to an Excel workbook.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2024, Brian Jacobs, brian.andrew.jacobs@gmail.com
#

import argparse
import ctypes
from pathlib import Path
import sys
import xlsxwriter


##################################################################
##################################################################
#
# WINLOGSTATS PDB FILE FORMAT
#
##################################################################
#
# 0|PLAY_DATA|0|PLAY_DATA|...|0|PLAY_DATA|1|TENDENCY_DATA|...|1|TENDENCY_DATA|<EOF>
#
# There are 0-N PLAY_DATA sections (one or more for each team in the file)
# There is 0-N TENDENCY_DATA sections (one for each team in the file)
# Occasionally, there is a zero-filled TENDENCY_DATA section
#
##################################################################
##################################################################

###############################################################################
#
# PLAY_DATA - A ctypes struct for storing play data
#
###############################################################################

class PLAY_DATA(ctypes.LittleEndianStructure):
    play_types = ["Run", "Pass", "Special", "Defense"]

    _fields_ = [("play_type",           ctypes.c_uint32),    # 0x00  // 0x00000000 // Run
                                                             #       // 0x00000001 // Pass
                                                             #       // 0x00000002 // Special
                                                             #       // 0x00000003 // Defense
                ("team_name",           ctypes.c_char*64),   # 0x04
                ("play_name",           ctypes.c_char*128),  # 0x44
                ("total_yards",         ctypes.c_int32),     # 0xC4  // OFFENSE
                ("play_count",          ctypes.c_uint32),    # 0xC8  // OFFENSE/DEFENSE/SPECIAL
                ("completions",         ctypes.c_uint32),    # 0xCC  // OFFENSE
                ("sacks",               ctypes.c_uint32),    # 0xD0	 // OFFENSE/DEFENSE
                ("fumbles",             ctypes.c_uint32),    # 0xD4	 // OFFENSE/DEFENSE
                ("interceptions",       ctypes.c_uint32),    # 0xD8	 // OFFENSE/DEFENSE
                ("touchdowns_offense",  ctypes.c_uint32),    # 0xDC	 // OFFENSE/DEFENSE
                ("touchdowns_defense",  ctypes.c_uint32),    # 0xE0	 // OFFENSE/DEFENSE
                ("reserved_1",          ctypes.c_int32),     # 0xE4	 // 0x00000000   // UNKNOWN
                ("reserved_2",          ctypes.c_int32),     # 0xE8	 // 0x00000000   // UNKNOWN
                ("points_scored",       ctypes.c_uint32),    # 0xEC	 // OFFENSE (TD points scored by offense or defense)
                                                             #       // DOESN'T COUNT TWO POINT CONVERSIONS
                ("run_plays_against",   ctypes.c_uint32),    # 0xF0	 // DEFENSE
                ("pass_plays_against",  ctypes.c_uint32),    # 0xF4	 // DEFENSE
                ("rush_yards_allowed",  ctypes.c_int32),     # 0xF8  // DEFENSE
                ("pass_yards_allowed",  ctypes.c_int32)]     # 0xFC  // DEFENSE

    def is_valid(self):
        # sanity check
        # check for valid play type, team name, and play name
        if (0 <= self.play_type < len(self.play_types)
                and len(self.team_name) > 0
                and len(self.play_name) > 0):
            return True
        return False


###############################################################################
#
# DOWN_DATA - A ctypes struct for storing situational data by down
#
###############################################################################

class DOWN_DATA(ctypes.LittleEndianStructure):
    _fields_ = [("first_down",  ctypes.c_uint32),           # 0x00
                ("second_down", ctypes.c_uint32),           # 0x04
                ("third_down",  ctypes.c_uint32),           # 0x08
                ("fourth_down", ctypes.c_uint32)]           # 0x0C


###############################################################################
#
# TENDENCY_DATA - A ctypes struct for storing tendency data by situatation
#
###############################################################################

class TENDENCY_DATA(ctypes.LittleEndianStructure):
    _fields_ = [("team_name",           ctypes.c_char*64),  # 0x00
                ("run_zero_to_one",     DOWN_DATA),         # 0x40
                ("pass_zero_to_one",    DOWN_DATA),         # 0x50
                ("run_two_to_five",     DOWN_DATA),         # 0x60
                ("pass_two_to_five",    DOWN_DATA),         # 0x70
                ("run_six_to_ten",      DOWN_DATA),         # 0x80
                ("pass_six_to_ten",     DOWN_DATA),         # 0x90
                ("run_ten_plus",        DOWN_DATA),         # 0xA0
                ("pass_ten_plus",       DOWN_DATA)]         # 0xB0

    def is_valid(self):
        return len(self.team_name) > 0


###############################################################################
#
# InvalidPDBError - A custom exception
#
###############################################################################

class InvalidPDBError(BaseException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


###############################################################################
#
# PDB - A class for opening and parsing a WinLogStats PDB file
#
###############################################################################

class PDB:
    data_types = ["play", "tendency"]

    ###########################################################################
    #
    # Initialization/Destruction
    #
    ###########################################################################

    def __init__(self, file_path):
        self.plays = []
        self.tendencies = []

        with open(file_path, 'rb') as pdb:
            # loop to process the entire file
            while True:
                # Read data type identifier of the next data chunk
                data = pdb.read(1)
                if not data:
                    # EOF
                    self.plays.sort(key=lambda x: x.team_name)
                    self.tendencies.sort(key=lambda x: x.team_name)
                    break
                data_type = int.from_bytes(data, byteorder='little')
                if data_type < 0 or data_type >= len(self.data_types):
                    raise InvalidPDBError(f"Invalid data type \'{data}\' at {pdb.tell()-1:#x} in {file_path}")
                if data_type == 0:
                    # Play data
                    data = pdb.read(ctypes.sizeof(PLAY_DATA))
                    play_data = PLAY_DATA.from_buffer_copy(data)
                    if not play_data.is_valid():
                        print(f"Skipping invalid play data at {pdb.tell()-ctypes.sizeof(PLAY_DATA):#x}")
                    else:
                        self.plays.append(play_data)
                elif data_type == 1:
                    # Tendency data
                    data = pdb.read(ctypes.sizeof(TENDENCY_DATA))
                    tendency_data = TENDENCY_DATA.from_buffer_copy(data)
                    if not tendency_data.is_valid():
                        print(f"Skipping invalid tendency data at {pdb.tell()-ctypes.sizeof(TENDENCY_DATA):#x}")
                    else:
                        self.tendencies.append(tendency_data)


###############################################################################
#
# PDBWorkbook - A class for creating the Excel workbook
#
###############################################################################

class PDBWorkbook:

    ###########################################################################
    #
    # Initialization/Destruction
    #
    ###########################################################################

    def __init__(self, file_path):
        self.workbook = xlsxwriter.Workbook(file_path)

        text_format = self.workbook.add_format({'num_format': '@'})
        avg_format = self.workbook.add_format({'num_format': '0.0'})

        # Prepare the 'Run Plays' worksheet
        self.run_rows = 0
        self.run_worksheet = self.workbook.add_worksheet('Run Plays')
        header = ['Team', 'Play', 'Calls', 'Yards', 'Avg', 'Fumbles', 'TD', 'Pts']
        self.run_worksheet.write_row(self.run_rows, 0, header)
        self.run_rows += 1
        self.run_worksheet.set_column_pixels('A:A', 140)
        self.run_worksheet.set_column_pixels('B:B', 140)
        self.run_worksheet.set_column('E:E', None, avg_format)
        self.run_worksheet.autofilter('A1:A1')
        self.run_worksheet.freeze_panes(1, 0)

        # Prepare the 'Pass Plays' worksheet
        self.pass_rows = 0
        self.pass_worksheet = self.workbook.add_worksheet('Pass Plays')
        header = ['Team', 'Play', 'Cmp/Att', 'Yards', 'Avg', 'Fumbles', 'Int', 'Sacks', 'TD', 'Pts']
        self.pass_worksheet.write_row(self.pass_rows, 0, header)
        self.pass_rows += 1
        self.pass_worksheet.set_column_pixels('A:A', 140)
        self.pass_worksheet.set_column_pixels('B:B', 140)
        self.pass_worksheet.set_column('C:C', None, text_format)
        self.pass_worksheet.set_column('E:E', None, avg_format)
        self.pass_worksheet.autofilter('A1:A1')
        self.pass_worksheet.freeze_panes(1, 0)

        # Prepare the 'Def Plays' worksheet
        self.def_rows = 0
        self.def_worksheet = self.workbook.add_worksheet('Def Plays')
        header = ['Team', 'Play', 'vs Run', 'Avg', 'vs Pass', 'Avg', 'Fumbles', 'Int', 'Sacks', 'TD/Def', 'TD/Off']
        self.def_worksheet.write_row(self.def_rows, 0, header)
        self.def_rows += 1
        self.def_worksheet.set_column_pixels('A:A', 140)
        self.def_worksheet.set_column_pixels('B:B', 140)
        self.def_worksheet.set_column('C:C', None, text_format)
        self.def_worksheet.set_column('D:D', None, avg_format)
        self.def_worksheet.set_column('E:E', None, text_format)
        self.def_worksheet.set_column('F:F', None, avg_format)
        self.def_worksheet.autofilter('A1:A1')
        self.def_worksheet.freeze_panes(1, 0)

        # Prepare the 'Tendencies' worksheet
        self.tendencies_rows = 0
        self.tendencies_worksheet = self.workbook.add_worksheet('Tendencies')
        header = ['Team', 'Situation', 'Runs', 'Passes']
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, header)
        self.tendencies_rows += 1
        self.tendencies_worksheet.set_column_pixels('A:A', 140)
        self.tendencies_worksheet.set_column_pixels('B:B', 140)
        self.tendencies_worksheet.autofilter('A1:A1')
        self.tendencies_worksheet.freeze_panes(1, 0)

    def __del__(self):
        if self.workbook:
            self.workbook.close()

    ###########################################################################
    #
    # Public API.
    #
    ###########################################################################

    def add_play(self, play_data):
        if PLAY_DATA.play_types[play_data.play_type] == 'Run':
            self._add_run_play(play_data)
        elif PLAY_DATA.play_types[play_data.play_type] == 'Pass':
            self._add_pass_play(play_data)
        elif PLAY_DATA.play_types[play_data.play_type] == 'Defense':
            self._add_defense_play(play_data)
        else:
            # Skip SPECIAL or ONSIDE
            pass

    def add_tendency(self, t):
        # First down
        row_data = [t.team_name.decode('ASCII'), 'First and 0-1', t.run_zero_to_one.first_down, t.pass_zero_to_one.first_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'First and 2-5', t.run_two_to_five.first_down, t.pass_two_to_five.first_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'First and 6-10', t.run_six_to_ten.first_down, t.pass_six_to_ten.first_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'First and 10+', t.run_ten_plus.first_down, t.pass_ten_plus.first_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1

        # Second down
        row_data = [t.team_name.decode('ASCII'), 'Second and 0-1', t.run_zero_to_one.second_down, t.pass_zero_to_one.second_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Second and 2-5', t.run_two_to_five.second_down, t.pass_two_to_five.second_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Second and 6-10', t.run_six_to_ten.second_down, t.pass_six_to_ten.second_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Second and 10+', t.run_ten_plus.second_down, t.pass_ten_plus.second_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1

        # Third down
        row_data = [t.team_name.decode('ASCII'), 'Third and 0-1', t.run_zero_to_one.third_down, t.pass_zero_to_one.third_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Third and 2-5', t.run_two_to_five.third_down, t.pass_two_to_five.third_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Third and 6-10', t.run_six_to_ten.third_down, t.pass_six_to_ten.third_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Third and 10+', t.run_ten_plus.third_down, t.pass_ten_plus.third_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1

        # Fourth down
        row_data = [t.team_name.decode('ASCII'), 'Fourth and 0-1', t.run_zero_to_one.fourth_down, t.pass_zero_to_one.fourth_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Fourth and 2-5', t.run_two_to_five.fourth_down, t.pass_two_to_five.fourth_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Fourth and 6-10', t.run_six_to_ten.fourth_down, t.pass_six_to_ten.fourth_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1
        row_data = [t.team_name.decode('ASCII'), 'Fourth and 10+', t.run_ten_plus.fourth_down, t.pass_ten_plus.fourth_down]
        self.tendencies_worksheet.write_row(self.tendencies_rows, 0, row_data)
        self.tendencies_rows += 1

    ###########################################################################
    #
    # Private API.
    #
    ###########################################################################

    def _add_run_play(self, play_data):
        if play_data.play_count > 0:
            avg = play_data.total_yards / play_data.play_count
        else:
            avg = 0.0

        row_data = [play_data.team_name.decode('ASCII'),
                    play_data.play_name.decode('ASCII'),
                    play_data.play_count,
                    play_data.total_yards,
                    round(avg, 1),
                    play_data.fumbles,
                    play_data.touchdowns_offense,
                    play_data.touchdowns_offense*6]

        self.run_worksheet.write_row(self.run_rows, 0, row_data)
        self.run_rows += 1

    def _add_pass_play(self, play_data):
        if play_data.completions > 0:
            avg = play_data.total_yards / play_data.completions
        else:
            avg = 0.0

        row_data = [play_data.team_name.decode('ASCII'),
                    play_data.play_name.decode('ASCII'),
                    f"{str(play_data.completions)}/{str(play_data.play_count)}",
                    play_data.total_yards,
                    round(avg, 1),
                    play_data.fumbles,
                    play_data.interceptions,
                    play_data.sacks,
                    play_data.touchdowns_offense,
                    play_data.touchdowns_offense*6]

        self.pass_worksheet.write_row(self.pass_rows, 0, row_data)
        self.pass_rows += 1

    def _add_defense_play(self, play_data):
        if play_data.run_plays_against > 0:
            rush_avg = play_data.rush_yards_allowed / play_data.run_plays_against
        else:
            rush_avg = 0.0

        if play_data.pass_plays_against > 0:
            pass_avg = play_data.pass_yards_allowed / play_data.pass_plays_against
        else:
            pass_avg = 0.0

        row_data = [play_data.team_name.decode('ASCII'),
                    play_data.play_name.decode('ASCII'),
                    f"{str(play_data.run_plays_against)}/{str(play_data.rush_yards_allowed)}",
                    round(rush_avg, 1),
                    f"{str(play_data.pass_plays_against)}/{str(play_data.pass_yards_allowed)}",
                    round(pass_avg, 1),
                    play_data.fumbles,
                    play_data.interceptions,
                    play_data.sacks,
                    play_data.touchdowns_defense,
                    play_data.touchdowns_offense]

        self.def_worksheet.write_row(self.def_rows, 0, row_data)
        self.def_rows += 1


def transform_pdb_to_xlsx(pdb_file_name, xlsx_file_name):
    try:
        print(f"Attempting to convert \'{pdb_file_name}\' into \'{xlsx_file_name}\'")

        pdb_path = Path(pdb_file_name)
        xlsx_path = Path(xlsx_file_name)
        xlsx_path.parent.mkdir(exist_ok=True)

        # Read the PDB
        pdb = PDB(pdb_path)

        # Create the XLSX file
        xlsx = PDBWorkbook(xlsx_path)

        # Add all play data in the PDB to the workbook
        for row_num, play_data in enumerate(pdb.plays):
            xlsx.add_play(play_data)

        # Add all tendency data in the PDB to the workbook
        for row_num, tendency_data in enumerate(pdb.tendencies):
            xlsx.add_tendency(tendency_data)

        print("Conversion complete")
        sys.exit(0)

    except InvalidPDBError as pe:
        print(f"{type(pe).__name__}: {pe}", file=sys.stderr)

    except Exception as e:
        print(f"{type(e).__name__} at line {e.__traceback__.tb_lineno} of {__file__}: {e}", file=sys.stderr)


###########################################################################
#
# MAIN ROUTINE
#
###########################################################################

def run(args=None):
    # Execute as command line script
    if not args:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "inputfile", help="read from this PDB file",
    )
    parser.add_argument(
        "outputfile", help="save to this XLSX file",
    )
    args = parser.parse_args(args)

    transform_pdb_to_xlsx(args.inputfile, args.outputfile)


if __name__ == "__main__":
    run()
